package com.alterdekim.flash.decompiler.translator;

import com.alterdekim.flash.decompiler.action.ActionPush;
import com.alterdekim.flash.decompiler.item.ActionPushType;
import com.alterdekim.flash.decompiler.action.ByteCodeAction;
import com.alterdekim.flash.decompiler.item.ByteCodeItem;
import com.alterdekim.flash.decompiler.item.PoolIndexItem;
import com.alterdekim.flash.decompiler.item.StringItem;
import com.alterdekim.flash.decompiler.tag.DoAction;
import com.alterdekim.flash.decompiler.util.ActionField;
import lombok.extern.slf4j.Slf4j;

import java.util.*;

@Slf4j
public class Flash2Java {

    private List<String> pool;

    public Map<String, FlashVar> convert(DoAction doAction) {
        this.pool = doAction.getActionPool();
        List<ByteCodeAction> actions = doAction.getActions();

        // convert stack from List<ByteCodeAction> to List<ByteCodeItem> and rewrite all code accordingly.
        List<ByteCodeAction> stack = new ArrayList<>();
        Map<String, FlashVar> vars = new HashMap<>();
        for (ByteCodeAction action : actions) {
            try {
                switch (action.getType()) {
                    case DefineLocal:
                        defineLocal(vars, stack);
                        stack.clear();
                        break;
                    case GetVariable:
                        getVariable(vars, stack);
                        stack.clear();
                        break;
                    case SetMember:
                        setMember(vars, stack);
                        stack.clear();
                        break;
                    case Add:
                        reduceViaAdder(stack);
                        break;
                    default:
                        stack.add(action);
                }
            } catch (Exception e) {
                log.error(e.getMessage());
            }
        }
        return vars;
    }

    private void reduceViaAdder(List<ByteCodeAction> stack) {
        // get last 2 elements from stack
        List<Integer> si = new ArrayList<>();
        List<ByteCodeItem> g = new ArrayList<>();
        for( int i = stack.size()-1; i >= 0; i-- ) {
            ByteCodeAction e = stack.get(i);
            if( e.getType() != ActionField.Push ) continue;
            ActionPush push = (ActionPush) e;
            if( push.getItems().size() >= 2 ) {
                StringItem from = (StringItem) push.getItems().get(push.getItems().size()-1);
                StringItem to = (StringItem) push.getItems().get(push.getItems().size()-2);
                push.getItems().remove()
                ((String) to.getValue()) + ((String) from.getValue())
            }
        }

    }

    private void setMember(Map<String, FlashVar> vars, List<ByteCodeAction> stack) {
        String varKey = vars.keySet().stream().filter(k -> vars.get(k).isActive()).findFirst().get();
        FlashVar varVal = vars.get(varKey);
        ActionPush push = (ActionPush) stack.get(0);
        if( stack.size() == 1 ) {
            Map<Integer, String> m = varVal.getVal() != null ? ((Map<Integer, String>) varVal.getVal()) : new HashMap<>();
            m.put((Integer) push.getItems().get(0).getValue(), fromPool((Integer) push.getItems().get(1).getValue()));
            varVal.setVal(m);
            vars.put(varKey, varVal);
            return;
        }
        Map<Integer, Map<String, Object>> m = varVal.getVal() != null ? ((Map<Integer, Map<String, Object>>) varVal.getVal()) : new HashMap<>();
        Map<String, Object> v = new HashMap<>();
        for( int i = 1; i < push.getItems().size()-1; i+=2 ) {
            v.put(fromPool((Integer) push.getItems().get(i).getValue()), valConverter(push.getItems().get(i+1)));
        }
        m.put((Integer) push.getItems().get(0).getValue(), v);
        varVal.setVal(m);
        vars.put(varKey, varVal);
    }

    private Object valConverter(ByteCodeItem item) {
        if ( item.getType() == ActionPushType.POOL_INDEX ) {
            return fromPool( (Integer) item.getValue() );
        } else if ( item.getType() == ActionPushType.STRING ) {
            return item.getValue();
        }
        return item.getValue();
    }

    private void getVariable(Map<String, FlashVar> vars, List<ByteCodeAction> stack) throws Exception {
        if (stack.size() != 1 ) { throw new Exception("Bad getVariable format"); }
        ActionPush push = (ActionPush) stack.get(0);
        PoolIndexItem pi = (PoolIndexItem) push.getItems().get(0);
        vars.keySet().forEach(k -> {
            FlashVar fv = vars.get(k);
            fv.setActive(false);
            vars.put(k, fv);
        });
        String key = fromPool((int) pi.getValue());
        FlashVar v = vars.get(key);
        v.setActive(true);
        vars.put(key, v);
    }

    private void defineLocal(Map<String, FlashVar> vars, List<ByteCodeAction> stack) throws Exception {
        if (stack.size() != 2 || stack.get(1).getType() != ActionField.InitObject) { throw new Exception("Bad defineLocal format"); }
        ActionPush push = (ActionPush) stack.get(0);
        PoolIndexItem pi = (PoolIndexItem) push.getItems().get(0);
        // todo: implement assignment ability
        vars.put(fromPool((Integer) pi.getValue()), new FlashVar());
    }

    private String fromPool(int index) {
        return pool.get(index);
    }
}
